#!/usr/bin/env python3
import common

import argparse
import logging
import sys
import os
from gitlab import Gitlab
from accept_merge_request import accept_merge_request
from create_merge_request import create_merge_request
from get_merge_requests import get_merge_requests
from update_submodule import update_submodule
from update_gitlab_ci import update_gitlab_ci_include

from ruamel.yaml import YAML


def create_gitlab_ci_yml(repo, gitlab_ci_yml):
    """This code snippet was used to initially populate
    all repos with the minial .gitlab-ci.yml that
    includes the files from this repo.
    Currently it is not used and only the revision is
    changed.
    It has also never been test as this function.
    """
    logging.debug("Origin: %s", repo.remotes.origin.url)

    # Contains the base file to be used in the subprojects
    if "ci-test" in repo.remotes.origin.url:
        if "manifest" in repo.remotes.origin.url:
            include_file = "foobar-manifest.yml"
        else:
            include_file = "foobar-manifest-integration.yml"
    else:
        if "manifest" in repo.remotes.origin.url:
            include_file = "manifest.yml"
        else:
            include_file = "manifest-integration.yml"

    # -----------------------------------
    # Adapt content in the include file
    # -----------------------------------
    # Read values from existing file
    yaml = YAML()
    with open(gitlab_ci_yml, "r", encoding="utf8") as fp:
        data = yaml.load(fp)

    logging.debug("Yaml: %s", data)

    try:
        masterbranch = data["variables"]["MASTER_BRANCH_PROJECT"]
        logging.debug("Masterbranch %s", masterbranch)
    except KeyError:
        masterbranch = None
    try:
        recipe = data["variables"]["BB_RECIPE_NAME"]
        logging.debug("Recipe %s", recipe)
    except KeyError:
        recipe = None

    # -----------------------------------
    with open(gitlab_ci_yml, "w", encoding="utf8") as fp:
        fp.write(
            """---
# ---------------------------------------------------------------------------------------
# Include the default CI steps from the gitlab-ci repo
# ---------------------------------------------------------------------------------------
include:
  - project: '${CI_PROJECT_ROOT_NAMESPACE}/yocto/infrastructure/gitlab-ci'
    ref: 49cc4204323bca76190e5ffae1a7d5627157c073
"""
        )
        fp.write("    file: '{}'\n".format(include_file))
        if masterbranch is not None or recipe is not None:
            fp.write("\nvariables:\n")
            if masterbranch is not None:
                fp.write("  MASTER_BRANCH_PROJECT: {}\n".format(masterbranch))
            if recipe is not None:
                fp.write("  BB_RECIPE_NAME: {}\n".format(recipe))

    # -----------------------------------
    repo.git.add(gitlab_ci_yml)

    # ======================================


def update_rev_in_gitlab_ci(repo, submodule_project, submodule_revision):

    # Add changed revision also to .gitlab-ci.yml
    gitlab_ci_yml = os.path.join(repo.working_tree_dir, ".gitlab-ci.yml")

    if update_gitlab_ci_include(
        gitlab_ci_yml,
        submodule_project.web_url.split("//")[1].split("/", 1)[1],
        submodule_revision,
    ):
        repo.git.add(gitlab_ci_yml)

    with open(gitlab_ci_yml, "r", encoding="utf8") as fp:
        logging.debug(fp.read())


def deploy_into(project, submodule, revision, branch, replace_existing_branch=False):
    """Update the submodule and include refs to the submodule in the given project.
        Create mergerequest if needed.

    Parameters:
        project ( gitlab project): The project which's submodule should be updated
        submodule_name (string): The name of the submodule to pull
        submodule_revision (hex string): The sha hash of the commit to update the submodule to
        branch (string): branch to update, if None, the projects default branch is used
        replace_existing_branch: When an existing integration branch is found it is always replaced.

    Returns: tuple of:
        branch (string): Name of the newly created integration branch
        merge_request (gitlab mr): Mergerequest for the integration branch
    """
    # Update submodule
    integration_branch, _, submodule_project = update_submodule(
        project,
        submodule,
        revision,
        branch,
        pre_commit_hook=update_rev_in_gitlab_ci,
        replace_existing_branch=replace_existing_branch,
    )

    logging.debug("Integration branch: %s", integration_branch)

    # If submodule is already at specified revision, return directly
    if not integration_branch:
        return None, submodule_project

    # Get source merge request
    mrs = get_merge_requests(
        submodule_project,
        # TODO should this be submodule_project's default branch?
        target_branch="master",
        commit=revision,
    )
    if not mrs:
        sys.exit(
            "ERROR: could not determine source merge request for commit %s" % revision
        )
    source_mr = mrs[0]

    # Create merge request
    mr, created = create_merge_request(
        project, integration_branch, project.default_branch
    )
    if created:
        common.crosslink_merge_requests(source_mr, mr)
        print("Created new merge request:\n%s" % mr.web_url)
    else:
        print("Existing integration merge request:\n%s" % mr.web_url)
    return integration_branch, mr


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--gitlab-url",
        help="""URL to the GitLab instance""",
        dest="gitlab_url",
        required=True,
    )
    parser.add_argument(
        "--token",
        help="""GitLab REST API private access token""",
        dest="token",
        required=True,
    )
    parser.add_argument(
        "--project",
        "--manifest-project",
        help="""name of the GitLab project""",
        dest="project",
    )
    parser.add_argument(
        "--submodule",
        help="""submodule to update""",
        dest="submodule",
        required=True,
    )
    parser.add_argument(
        "--revision",
        help="""new revision for submodule""",
        dest="revision",
        required=True,
    )
    parser.add_argument(
        "--branch",
        help="""project branch (if not default branch)""",
        dest="branch",
        required=False,
        default=None,
    )
    parser.add_argument(
        "--merge",
        help="""if set, perform merge after integration""",
        dest="merge",
        action="store_true",
        required=False,
        default=False,
    )
    parser.add_argument(
        "projects",
        help="""List of projects the change should be deployed to additionally
                to the manifest project given as named parameter.""",
        nargs="*",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="""Increase verbosity.""",
    )

    args, _ = parser.parse_known_args()
    if args.verbose:
        logging.basicConfig(level=logging.DEBUG)

    gitlab = Gitlab(args.gitlab_url, private_token=args.token)
    project = common.get_project(gitlab, args.project)

    logging.debug("Integrate into: %s", args.project)

    # Update submodule in this project, create MR
    integration_branch, mr = deploy_into(
        project,
        args.submodule,
        args.revision,
        args.branch,
        replace_existing_branch=len(args.projects) > 0,
    )
    merge_request_manifest = mr
    merge_requests = []

    # If submodule is already at specified revision, exit successfully
    if not integration_branch:
        sys.exit(0)

    for p in args.projects:
        gitlab_p = common.get_project(gitlab, p)
        logging.debug("Integrate into: %s", p)

        integration_branch, mr = deploy_into(
            gitlab_p,
            args.submodule,
            args.revision,
            args.branch,
        )
        merge_requests.append(mr)

        logging.debug("Integration branch: %s", integration_branch)

    if not args.merge:
        print(
            "Skipping automatic merge in MR context. If you like to extend the "
            "integration MR by hand, please do it now. Afterwards you can either merge "
            "it by hand or re-run this job on the master branch after the source MR "
            "has been merged."
        )
        sys.exit(0)

    # The manifest needs to be merged at last
    merge_requests.append(merge_request_manifest)
    for mr in merge_requests:
        logging.debug("Merge %s", mr)
        # Wait until GitLab has checked merge status
        common.wait_until_merge_status_is_set(project, mr)

        # Attempt to merge
        merged = accept_merge_request(project, mr, rebase=True)

        if not merged:
            sys.exit(
                "Integration MR could not be merged. You have two possibilities to fix "
                "this:\n"
                "  1. Checkout the MR and rebase it on the current master manually, or\n"
                "  2. Delete the MR (Edit -> Delete in the MR UI)\n"
                "In either case restart this job afterwards in order to get it merged."
            )

    print("Successfully merged")


if __name__ == "__main__":
    main()
