#!/usr/bin/env python3
import common

import argparse
import logging
import sys
from gitlab import Gitlab, GitlabGetError
from accept_merge_request import accept_merge_request
import update_submodule
from gitlab.v4.objects import Project


def find_integration_merge_request(
    project: Project, submodule, revision, target_branch
):

    (
        submodule_path,
        submodule_revision,
    ) = update_submodule.get_submodule_project_path_and_revision(
        project, submodule, target_branch
    )
    logging.debug("Module: %s, Revision: %s", submodule_path, submodule_revision)

    # Get submodule project
    gitlab = project.manager.gitlab
    submodule_project = common.get_project(gitlab, submodule_path)

    # Get the integration branch name
    integration_branch_suffix = (
        update_submodule.get_submodule_integration_branch_suffix(
            submodule_project, revision
        )
    )
    integration_branch_name = common.integration_branch_name(
        submodule_project.name, integration_branch_suffix
    )
    logging.debug(integration_branch_name)

    # Get the merge request for the branch TODO catch exception
    try:
        integration_branch = project.branches.get(
            integration_branch_name, retry_transient_errors=True
        )
    except GitlabGetError:
        print("ERROR: Failed to find integration branch for {}".format(submodule))
        return None

    logging.debug(integration_branch)
    commit = project.commits.get(
        integration_branch.commit["id"], retry_transient_errors=True
    )
    for mr in commit.merge_requests(retry_transient_errors=True):
        if mr["target_branch"] == target_branch:
            integration_mr = mr
            break
    if integration_mr is None:
        print(
            "ERROR: Failed to find integration merge request for %s",
            integration_branch_name,
        )
        return None
    logging.debug(integration_mr["iid"])
    mr = project.mergerequests.get(integration_mr["iid"], retry_transient_errors=True)

    return mr


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--gitlab-url",
        help="""URL to the GitLab instance""",
        dest="gitlab_url",
        required=True,
    )
    parser.add_argument(
        "--token",
        help="""GitLab REST API private access token""",
        dest="token",
        required=True,
    )
    parser.add_argument(
        "--project",
        "--manifest-project",
        help="""name of the GitLab project""",
        dest="project",
    )
    parser.add_argument(
        "--submodule",
        help="""submodule to update""",
        dest="submodule",
        required=True,
    )
    parser.add_argument(
        "--revision",
        help="""new revision for submodule""",
        dest="revision",
        required=True,
    )
    parser.add_argument(
        "--branch",
        help="""project branch (if not default branch)""",
        dest="branch",
        required=False,
        default=None,
    )
    parser.add_argument(
        "--merge",
        help="""if set, perform merge after integration""",
        dest="merge",
        action="store_true",
        required=False,
        default=False,
    )
    parser.add_argument(
        "projects",
        help="""List of projects the change should be deployed to additionally
                to the manifest project given as named parameter.""",
        nargs="*",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="""Increase verbosity.""",
    )

    args, _ = parser.parse_known_args()
    if args.verbose:
        logging.basicConfig(level=logging.DEBUG)

    gitlab = Gitlab(args.gitlab_url, private_token=args.token)

    # Start with the manifest here, so the subproject
    # can see that they are already integrated
    for p in [args.project] + args.projects:
        project = common.get_project(gitlab, p)
        branch = args.branch
        should_remove_source_branch = True

        if branch is None:
            branch = project.default_branch

        print(
            "Try to merge {}({}) into {} ({})".format(
                args.submodule, args.revision, project.name, branch
            )
        )

        mr = find_integration_merge_request(
            project, args.submodule, args.revision, branch
        )
        if mr is None:
            sys.exit("ERROR: Failed to find the integration MR.")
            # TODO if this ever happens, why ever, we could call
            # deploy_gitlab_ci again to create a new integration
            # commit or make sure the change is already integrated.

        # Store for later usage:
        if p == args.project:
            manifest_project = project
            manifest_integration_mr = mr
            # For the manifest we keep the source branch
            # otherwise does the check job fail for not yet merged
            # child projects
            should_remove_source_branch = False

        print("Merge {}!{}: {}".format(project.name, mr.iid, mr.title))
        # Wait until GitLab has checked merge status
        common.wait_until_merge_status_is_set(project, mr)

        # Attempt to merge
        merged = accept_merge_request(
            project,
            mr,
            # Only the file manifest commit may be rebased
            # other rebase would require new integration commits
            rebase=(p == args.project),
            should_remove_source_branch=should_remove_source_branch,
        )

        if not merged:
            print(
                "Integration MR could not be merged."
                "To fix this:\n"
                "  1. Checkout the MR {}!{} and merge it manually.\n"
                "  2. Manually merge the follow up MRs in the following project:".format(
                    project.name, mr.iid
                )
            )
            found = False
            for p2 in [args.project] + args.projects:
                if p2 == p:
                    found = True
                    continue
                if not found:
                    continue
                print("     {}".format(p2))
            sys.exit()

    # Delete args.project (manifests) integration branch
    # (skipped the deletion during the merge request merge before)
    integration_branch = manifest_project.branches.get(
        manifest_integration_mr.source_branch
    )
    manifest_project.branches.delete(
        integration_branch.get_id(), retry_transient_errors=True
    )

    print("Successfully merged")

    exit()


if __name__ == "__main__":
    main()
